# Copyright (C) 2023 Cardiff University
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

"""HTCondor submit command formatting utilities.
"""

from __future__ import annotations

from typing import Any, Callable, Iterable


def _separated_factory(delimiter: str) -> Callable:
    """Factory function to define a delimiter-joiner function."""

    def _joiner(value: Iterable[Any]) -> str:
        return delimiter.join(map(str, value))

    return _joiner


_comma_separated = _separated_factory(", ")
_space_separated = _separated_factory(" ")
_semicolon_separated = _separated_factory("; ")


def _format_argument(arg: str) -> str:
    """Format a single command argument for inclusion in the condor_submit
    ``arguments`` command.

    According to the condor_submit _New Syntax_ description.
    """
    return "'{}'".format(
        arg.replace('"', '""').replace("'", "''"),
    )


def _format_arguments(value: Iterable[str]) -> str:
    """Format an iterable of command arguments for inclusion in the condor_submit
    ``arguments`` command.

    According to the condor_submit _New Syntax_ description.
    """
    return '"{}"'.format(" ".join(map(_format_argument, value)))


def _format_environment(value: dict[str, Any]) -> str:
    """Format a `dict` of environment settings for the condor_submit
    ``environment`` command.
    """
    return '"{}"'.format(" ".join(f"{key}='{value}'" for key, value in value.items()))


def _format_classad_item(value: Any) -> str:
    """Format a single classad expression item."""
    return f"({value})"


def _format_classad_boolean_expression(
    value: Iterable[Any],
    operator: str = "&&",
) -> str:
    """Format an iterable of requirements expressions as a ClassAd Boolean
    Expression for the condor_submit ``requirements`` command, or similar,
    using the relevant operator.
    """
    return " {operator} ".join(map(_format_classad_item, value))


def _semicolon_separated_pairs(value: dict[str, Any]) -> str:
    """Format a dict as a semicolon-separated list of key, value pairs
    for the condor_submit ``transfer_output_remaps`` command, or similar.
    """

    def _format_item(key):
        val = value[key]
        if val is None:
            return key
        return f"{key}={val}"

    return '"{}"'.format("; ".join(map(_format_item, value)))


DICT_FORMATTERS: dict[str, Callable] = {
    "environment": _format_environment,
    "transfer_output_remaps": _semicolon_separated_pairs,
    "transfer_plugins": _semicolon_separated_pairs,
}
LIST_FORMATTERS: dict[str, Callable] = {
    "arguments": _format_arguments,
    "getenv": _comma_separated,
    "requirements": _format_classad_boolean_expression,
    "transfer_checkpoint_files": _comma_separated,
    "transfer_input_files": _comma_separated,
    "transfer_output_files": _comma_separated,
}


def _format_condor_command_value(
    command: str,
    value: Any,
) -> str:
    """Format a condor submit command value."""
    if isinstance(value, dict) and command.lower() in DICT_FORMATTERS:
        return DICT_FORMATTERS[command.lower()](value)
    if isinstance(value, (list, tuple)) and command.lower() in LIST_FORMATTERS:
        return LIST_FORMATTERS[command.lower()](value)

    # special cases for dynamically-named commands
    if isinstance(value, (list, tuple)) and (
        "_oauth_permissions" in command or "_oauth_resource" in command
    ):
        return _space_separated(value)

    return str(value)


def format_submit_description(desc: dict[str, Any]) -> dict[str, str]:
    """Format a dictionary of submit description options.

    This method converts 'arbitrary' Pythonic input types for the values of
    condor submit description options into a string for inclusion in a
    condor submit file.
    """
    out = {}
    for key, value in desc.items():
        out[key] = _format_condor_command_value(key, value)
    return out



# Copyright (C) 2022 Patrick Godwin
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

from ._version import version as __version__

from .dags import DAG as DAG
from .layers import (
    Layer as Layer,
    Node as Node,
)
from .options import (
    Argument as Argument,
    Option as Option,
)



# Copyright (C) 2023 Patrick Godwin, Cardiff University
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

import os

from typing import Callable, Union


def is_abs_or_url(path: str) -> bool:
    """Check whether a path is absolute or URL-based."""
    if os.path.isabs(path):
        return True
    return "://" in str(path)


def normalize(path: str, basename: Union[bool, Callable[[str], bool]] = False) -> str:
    """Selectively return the path's basename based on a condition."""
    if (callable(basename) and basename(path)) or basename is True:
        return os.path.basename(path)
    else:
        return path



# Copyright (C) 2020 Patrick Godwin
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

from collections import defaultdict
import os
from pathlib import Path
import re
from typing import Dict, List, Optional, Tuple
import warnings

# disable warnings when condor config source is not found
with warnings.catch_warnings():
    warnings.simplefilter("ignore", UserWarning)
    import htcondor
    from htcondor import dags

from .layers import HexFormatter, Layer


class DAG(dags.DAG):
    """Defines a DAGMan workflow including the execution graph and configuration.

    Parameters
    ----------
    name : str
        The name of the DAG workflow, used for files written to disk and for
        DAG submission when calling write() and submit(). Defaults to "workflow".
    formatter : htcondor.dags.NodeNameFormatter
        Defines how the node names are defined and formatted. Defaults to a
        hex-based formatter with 5 digits.
    *args
        Any positional arguments that htcondor.dags.DAG accepts
    **kwargs
        Any keyword arguments that htcondor.dags.DAG accepts

    """

    def __init__(
        self,
        name: str = "workflow",
        formatter: Optional[dags.NodeNameFormatter] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.name = name
        self._node_layers: Dict[str, dags.NodeLayer] = {}
        self._ordered_layers: List[dags.NodeLayer] = []
        self._layers: Dict[str, Layer] = {}
        self._provides: Dict[str, Tuple[str, int]] = {}
        if formatter:
            self.formatter = formatter
        else:
            self.formatter = HexFormatter()
        self._dag_path: Optional[str] = None

    def attach(self, layer: Layer) -> None:
        """Attach a layer of related job nodes to this DAG.

        Parameters
        ----------
        layer
            The layer to attach.

        """
        key = layer.name
        if key in self._layers:
            raise KeyError(f"{key} layer already added to DAG")
        self._layers[layer.name] = layer

        # determine parent-child relationships and connect accordingly
        all_edges = defaultdict(set)
        if layer.has_dependencies:
            # determine edges
            for child_idx, node in enumerate(layer.nodes):
                for input_ in node.requires:
                    if input_ in self._provides:
                        parent_name, parent_idx = self._provides[input_]
                        all_edges[parent_name].add((parent_idx, child_idx))

            if not all_edges:
                node_layer = self.layer(**layer.config(self.formatter))
                self._node_layers[key] = node_layer
                self._ordered_layers.append(node_layer)

            # determine edge type and connect
            for num, (parent, edges) in enumerate(all_edges.items()):
                edge = self._get_edge_type(parent, layer.name, edges)
                if num == 0:
                    node_layer = self._node_layers[parent].child_layer(
                        **layer.config(self.formatter), edge=edge
                    )
                    self._node_layers[key] = node_layer
                    self._ordered_layers.append(node_layer)
                else:
                    self._node_layers[key].add_parents(
                        self._node_layers[parent], edge=edge
                    )

        else:
            node_layer = self.layer(**layer.config(self.formatter))
            self._node_layers[key] = node_layer
            self._ordered_layers.append(node_layer)

        # register any data products the layer provides
        for idx, node in enumerate(layer.nodes):
            for output in node.provides:
                self._provides[output] = (key, idx)

    def create_log_dir(self, log_dir: Path = Path("logs")) -> None:
        """Create the log directory where job logs are stored.

        Parameters
        ----------
        log_dir : Path
            The directory to create logs in. Defaults to ./logs.

        """
        warnings.warn(
            "create_log_dir has been deprecated in favor of automatically "
            "creating log directories upon DAG generation. this method "
            "will be removed in a future release",
            DeprecationWarning,
        )
        os.makedirs(log_dir, exist_ok=True)

    def write(self, path: Path = Path.cwd(), write_script: bool = False) -> None:
        """Write out the given DAG to the given directory.

        This includes the DAG description file itself, as well as any
        associated submit descriptions and log directories.

        Also optionally writes out the list of commands for each node, which
        represents commands that would be run on the execute point, after
        taking into account file location changes where the job would be run if
        file transfer is enabled.

        Parameters
        ----------
        path : Path
            The directory to write the DAG files to. Defaults to the current working
            directory.
        write_script : bool
            Also write out the list of commands for each node to disk. Defaults
            to false.

        """
        dag_file = f"{self.name}.dag"
        self._write_dag(dag_file, path=path)
        self._dag_path = str(path / dag_file)
        if write_script:
            self._write_script(f"{self.name}.sh", path=path)

    def submit(
        self, path: Path = Path.cwd(), write_script: bool = False, **kwargs
    ) -> None:
        """Submit the DAG via HTCondor.

        If the DAG has not already been written to disk, do so as well.
        This is equivalent to calling write() prior to submission, making
        use of the `path` and `write_script` arguments for doing so. See
        DAG.write for more information.

        Parameters
        ----------
        path : Path
            The directory to write the DAG files to. Defaults to the current working
            directory.
        write_script : bool
            Also write out the list of commands for each node to disk. Defaults
            to false.
        **kwargs
            Any keyword arguments that `condor_submit_dag` accepts. See
            [htcondor.Submit.from_dag](https://htcondor.readthedocs.io/en/latest/apis/python-bindings/api/htcondor.html#htcondor.Submit.from_dag)
            for more information.

        """
        # write DAG to disk if not already done
        if not self._dag_path:
            self.write(path, write_script)
            self._dag_path = str(path / f"{self.name}.dag")

        # submit the DAG
        submit_kwargs = {"UseDagDir": True, **kwargs}
        dag_submit = htcondor.Submit.from_dag(self._dag_path, submit_kwargs)
        htcondor.Schedd().submit(dag_submit)

    def write_dag(self, filename: str, path: Path = Path.cwd(), **kwargs) -> None:
        """Write out the given DAG to the given directory.

        This includes the DAG description file itself, as well as any
        associated submit descriptions and log directories.

        Parameters
        ----------
        filename : str
            The name of the DAG description file itself, e.g. my_dag.dag.
        path : Path
            The directory to write the DAG files to. Defaults to the current working
            directory.
        **kwargs
            Any other keyword arguments that htcondor.dags.write_dag accepts

        """
        warnings.warn(
            "write_dag has been deprecated in favor of write. "
            "this method will be removed in a future release",
            DeprecationWarning,
        )
        self._write_dag(filename, path, **kwargs)

    def write_script(
        self,
        filename: str,
        path: Path = Path.cwd(),
    ) -> None:
        """Write out the list of commands for each node to the given directory.

        This represents commands that would be run on the execute point, after
        taking into account file location changes where the job would be run if
        file transfer is enabled.

        Parameters
        ----------
        filename : str
            The name of the script file itself, e.g. my_dag.sh.
        path : Path
            The directory to write the script file to. Defaults to the current working
            directory.

        """
        warnings.warn(
            "write_dag has been deprecated in favor of write. "
            "this method will be removed in a future release",
            DeprecationWarning,
        )
        self._write_script(filename, path)

    def _write_dag(self, filename: str, path: Path = Path.cwd(), **kwargs) -> None:
        """Write out the given DAG to the given directory.

        This includes the DAG description file itself, as well as any
        associated submit descriptions and log directories.

        Parameters
        ----------
        filename : str
            The name of the DAG description file itself, e.g. my_dag.dag.
        path : Path
            The directory to write the DAG files to. Defaults to the current working
            directory.
        **kwargs
            Any other keyword arguments that htcondor.dags.write_dag accepts

        """
        # create log directories
        for name, layer in self._layers.items():
            if os.path.isabs(layer.log_dir):
                log_path = Path(layer.log_dir)
            else:
                submit = self._node_layers[name].submit_description
                initialdir = Path(submit.get("initialdir", str(path)))
                log_path = initialdir / layer.log_dir
            os.makedirs(log_path, exist_ok=True)

        # create DAG and submit files
        dags.write_dag(
            self,
            path,
            dag_file_name=filename,
            node_name_formatter=self.formatter,
            **kwargs,
        )

    def _write_script(
        self,
        filename: str,
        path: Path = Path.cwd(),
    ) -> None:
        """Write out the list of commands for each node to the given directory.

        This represents commands that would be run on the execute point, after
        taking into account file location changes where the job would be run if
        file transfer is enabled.

        Parameters
        ----------
        filename : str
            The name of the script file itself, e.g. my_dag.sh.
        path : Path
            The directory to write the script file to. Defaults to the current working
            directory.

        """
        with open(path / filename, "w") as f:
            # traverse DAG in breadth-first order
            for layer in self._ordered_layers:
                # grab relevant submit args, format $(arg) to {arg}
                executable = layer.submit_description["executable"]
                args = layer.submit_description["arguments"]
                args = re.sub(r"\$\(((\w+?))\)", r"{\1}", args)

                # evaluate vars for each node in layer, write to disk
                for idx, node_vars in enumerate(layer.vars):
                    node_name = self.formatter.generate(layer.name, idx)
                    print(f"# Job {node_name}", file=f)
                    print(executable + " " + args.format(**node_vars) + "\n", file=f)

    def _get_edge_type(self, parent_name, child_name, edges) -> dags.BaseEdge:
        parent = self._layers[parent_name]
        child = self._layers[child_name]
        edges = sorted(list(edges))

        # check special cases, defaulting to explicit edge connections via indices
        if len(edges) == (len(parent.nodes) + len(child.nodes)):
            return dags.ManyToMany()

        elif len(parent.nodes) == len(child.nodes) and all(
            [parent_idx == child_idx for parent_idx, child_idx in edges]
        ):
            return dags.OneToOne()

        else:
            return EdgeConnector(edges)


class EdgeConnector(dags.BaseEdge):
    """This edge connects individual nodes in layers given an explicit mapping."""

    def __init__(self, indices) -> None:
        self.indices = indices

    def get_edges(self, parent, child, join_factory):
        for parent_idx, child_idx in self.indices:
            yield (parent_idx,), (child_idx,)


def write_dag(
    dag: dags.DAG,
    dag_dir: Path = Path.cwd(),
    formatter: Optional[dags.NodeNameFormatter] = None,
    **kwargs,
) -> Path:
    """Write out the given DAG to the given directory.

    This includes the DAG description file itself, as well as any associated
    submit descriptions.

    Parameters
    ----------
    dag : DAG
        The DAG to write.
    dag_dir : Path
        The directory to write the DAG files to. Defaults to the current working
        directory.
    formatter : htcondor.dags.NodeNameFormatter
        Defines how the node names are defined and formatted. Defaults to a
        hex-based formatter with 5 digits.
    **kwargs
        Any other keyword arguments that htcondor.dags.write_dag accepts

    """
    warnings.warn(
        "write_dag has been deprecated in favor of DAG.write. "
        "this method will be removed in a future release",
        DeprecationWarning,
    )
    if not formatter:
        formatter = HexFormatter()
    return dags.write_dag(dag, dag_dir, node_name_formatter=formatter, **kwargs)



import pytest

from ..options import Argument, Literal, Option
from .. import path


@pytest.mark.parametrize(
    "name,value,is_file",
    [
        ["index", 1, False],
        ["format", "csv", False],
        ["input", "file1.txt", True],
        ["input", "to/file1.txt", True],
        ["input", "/path/to/file1.txt", True],
    ],
)
def test_argument_commands_single(name, value, is_file):
    argument = Argument(name, value)

    assert argument.name == name
    assert argument.vars() == str(value)

    if is_file:
        assert argument.files() == value

        filename = path.normalize(value, basename=path.is_abs_or_url)
        assert argument.files(basename=path.is_abs_or_url) == filename

        if filename == value:
            assert not argument.remaps()
        else:
            assert argument.remaps() == f"{filename}={value}"


def test_option_commands_flag():
    name = "verbose"
    option = Option(name)

    assert option.name == name
    assert option.vars() == f"--{name}"


@pytest.mark.parametrize(
    "name,value,is_file",
    [
        ["index", 1, False],
        ["format", "csv", False],
        ["input", "file1.txt", True],
        ["input", "to/file1.txt", True],
        ["input", "/path/to/file1.txt", True],
    ],
)
def test_option_commands_single(name, value, is_file):
    option = Option(name, value)

    assert option.name == name
    assert option.vars() == f"--{name} {value}"

    if is_file:
        assert option.files() == value

        filename = path.normalize(value, basename=path.is_abs_or_url)
        assert option.files(basename=path.is_abs_or_url) == filename

        if filename == value:
            assert not option.remaps()
        else:
            assert option.remaps() == f"{filename}={value}"


@pytest.mark.parametrize(
    "value,is_file",
    [
        [1, False],
        ["csv", False],
        ["file1.txt", True],
        ["to/file1.txt", True],
        ["/path/to/file1.txt", True],
    ],
)
def test_literal_commands_single(value, is_file):
    literal = Literal(value)

    assert literal.vars() == str(value)

    if is_file:
        assert literal.files() == str(value)

        filename = path.normalize(value, basename=path.is_abs_or_url)
        assert literal.files(basename=path.is_abs_or_url) == filename

        if filename == value:
            assert not literal.remaps()
        else:
            assert literal.remaps() == f"{filename}={value}"



# Copyright (C) 2020 Patrick Godwin
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass, field
import itertools
import os
import re
import shutil
from typing import Any, Dict, List, Optional, Tuple, Union
import warnings

# disable warnings when condor config source is not found
with warnings.catch_warnings():
    warnings.simplefilter("ignore", UserWarning)
    import htcondor
    from htcondor import dags

from .format import format_submit_description
from .options import Argument, Literal, Option, PROTECTED_CONDOR_VARS
from . import path


@dataclass
class Layer:
    """Defines a single layer (or set of related jobs) in an HTCondor DAG.

    Stores submit configuration for a set of nodes as well as
    providing functionality to determine the parent-child
    relationships between nodes.

    Parameters
    ----------
    executable
        The path of the executable to run.
    name
        The human-readable name of this node. Defaults to the basename
        of the executable if not given.
    universe
        The execution environment for a job. Defaults to 'vanilla'.
    log_dir
        The directory in which logs will be written to. Defaults to ./logs.
    retries
        The number of retries given for a job. Defaults to 3.
    transfer_files
        Whether to leverage Condor file transfer for moving around
        files. On by default.
    submit_description
        The submit descriptors representing this set of jobs.
    requirements
        The submit descriptors representing this set of jobs.
        Deprecated in favor for submit_description to avoid confusion,
        as 'requirements' refers to a specific submit descriptor.
        This option will be removed in a future release.
    nodes
        The nodes representing the layer. Nodes can be passed upon
        instantiation or added to the layer after the fact via
        Layer.append(node), Layer.extend(nodes), or Layer += node.

    """

    executable: str
    name: str = ""
    universe: str = "vanilla"
    log_dir: str = "logs"
    retries: int = 3
    transfer_files: bool = True
    requirements: dict = field(default_factory=dict)
    submit_description: Union[dict, htcondor.Submit] = field(default_factory=dict)
    nodes: list = field(default_factory=list)
    inputs: dict = field(init=False, default_factory=dict)
    outputs: dict = field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        if not self.name:
            self.name = os.path.basename(self.executable)
        if self.requirements:
            self.submit_description.update(self.requirements)
            warnings.warn(
                "requirements has been deprecated in favor of submit_description"
                "to avoid confusion and will be removed in a future release",
                DeprecationWarning,
            )
        self.extend(self.nodes)

    def config(
        self,
        formatter: Optional[dags.NodeNameFormatter] = None,
    ) -> Dict[str, Any]:
        """Generates a layer configuration.

        This configuration can be passed directly into an
        htcondor.dags.NodeLayer if desired.

        Parameters
        ----------
        formatter : htcondor.dags.NodeNameFormatter
            Defines how the node names are defined and formatted. Defaults to a
            hex-based formatter with 5 digits.

        """
        # check that nodes are valid
        self.validate()

        # update submit description with defaults + other layer configuration
        submit_description = self._update_submit_defaults(self.submit_description)

        if not formatter:
            formatter = HexFormatter()
        return {
            "name": self.name,
            "submit_description": submit_description,
            "vars": self._vars(formatter),
            "retries": self.retries,
        }

    def append(self, node: Node) -> None:
        """Append a node to this layer."""
        assert isinstance(node.inputs, list)
        assert isinstance(node.outputs, list)
        for input_ in node.inputs:
            self.inputs.setdefault(input_.name, []).append(input_.argument)
        for output in node.outputs:
            self.outputs.setdefault(output.name, []).append(output.argument)
        self.nodes.append(node)

    def extend(self, nodes: Iterable[Node]) -> None:
        """Append multiple nodes to this layer."""
        for node in nodes:
            self.append(node)

    def __iadd__(self, nodes) -> Layer:
        if isinstance(nodes, Iterable):
            self.extend(nodes)
        else:
            self.append(nodes)
        return self

    def validate(self) -> None:
        """Ensure all nodes in this layer are consistent with each other."""
        assert self.nodes, "at least one node must be connected to this layer"

        # check arg names across nodes are equal
        args = [arg.name for arg in self.nodes[0].arguments]
        for node in self.nodes[:-1]:
            assert args == [arg.name for arg in node.arguments]

        # check input/output names across nodes are equal
        inputs = [arg.name for arg in self.nodes[0].inputs]
        for node in self.nodes[:-1]:
            assert inputs == [arg.name for arg in node.inputs]
        outputs = [arg.name for arg in self.nodes[0].outputs]
        for node in self.nodes[:-1]:
            assert outputs == [arg.name for arg in node.outputs]

        # check meta-parameters (equality, name validity)
        variables = list(self.nodes[0].variables.keys())
        for node in self.nodes[:-1]:
            assert variables == list(node.variables.keys())
        for var in variables:
            if var in PROTECTED_CONDOR_VARS:
                raise ValueError(
                    f"{var} is a protected condor name for node {self.name}"
                )

    def command(self, node, readjust_paths: bool = True):
        """Given a node, return the command that would be run.

        Parameters
        ----------
        node : Node
            The node to return the command for.
        readjust_paths : bool
            Determines whether path locations are readjusted based on
            the command that would be run on the node's execute point.
            This only has an effect if using file transfer. Default is True.

        """
        args = re.sub(r"\$\(((\w+?))\)", r"{\1}", self._arguments())
        # extract node variables
        node_vars = {arg.condor_name: arg.vars() for arg in node.arguments}
        for arg in node.inputs:
            if self.transfer_files and readjust_paths:
                node_vars[arg.condor_name] = arg.vars(basename=path.is_abs_or_url)
            else:
                node_vars[arg.condor_name] = arg.vars()
        for arg in node.outputs:
            basename = readjust_paths and self.transfer_files
            node_vars[arg.condor_name] = arg.vars(basename=basename)
        return self.executable + " " + args.format(**node_vars)

    @property
    def has_dependencies(self) -> bool:
        """Check if any of the nodes in this layer have dependencies."""
        return any([node.requires for node in self.nodes])

    def _arguments(self) -> str:
        args = []
        for arg in self.nodes[0].arguments:
            args.append(arg.vars() if arg.static else f"$({arg.condor_name})")
        io_args = []
        io_opts = []
        for arg in itertools.chain(self.nodes[0].inputs, self.nodes[0].outputs):
            if arg.static:
                basename = path.is_abs_or_url if self.transfer_files else False
                io_args.append(arg.vars(basename=basename))
            elif not arg.suppress:
                if isinstance(arg, Argument):
                    io_args.append(f"$({arg.condor_name})")
                else:
                    io_opts.append(f"$({arg.condor_name})")
        return " ".join(itertools.chain(args, io_opts, io_args))

    def _inputs(self) -> str:
        inputs = []
        for arg in self.nodes[0].inputs:
            if arg.static:
                inputs.append(arg.files())
            else:
                inputs.append(f"$(input_{arg.condor_name})")
        return ",".join(inputs)

    def _outputs(self) -> str:
        outputs = []
        for arg in self.nodes[0].outputs:
            if arg.static:
                outputs.append(arg.files(basename=path.is_abs_or_url))
            else:
                outputs.append(f"$(output_{arg.condor_name})")
        return ",".join(outputs)

    def _output_remaps(self) -> str:
        remaps = []
        for arg in self.nodes[0].outputs:
            if arg.static:
                remaps.append(arg.remaps())
            else:
                remaps.append(f"$(output_{arg.condor_name}_remap)")
        return ";".join(remaps)

    def _vars(self, formatter: dags.NodeNameFormatter) -> List[Dict[str, str]]:
        allvars = []
        for i, node in enumerate(self.nodes):
            nodevars = {
                "nodename": formatter.generate(self.name, i),
                "log_dir": self.log_dir,
                **node.variables,
            }

            # add arguments which aren't suppressed
            for arg in node.arguments:
                if not arg.static and not arg.suppress:
                    nodevars[arg.condor_name] = arg.vars()

            # then add arguments defined as 'inputs'. if file transfer is enabled,
            # also define the $(input_{arg}) variable containing the files
            for arg in node.inputs:
                if not arg.static:
                    if not arg.suppress:
                        basename = path.is_abs_or_url if self.transfer_files else False
                        nodevars[arg.condor_name] = arg.vars(basename=basename)
                    if self.transfer_files:
                        # adjust file location for input files if they are
                        # absolute paths. condor will transfer the file
                        # /path/to/file.txt to the job's current working
                        # directory, so arguments should point to file.txt
                        nodevars[f"input_{arg.condor_name}"] = arg.files()

            # finally, add arguments defined as 'outputs'. if file transfer is
            # enabled, also define the $(output_{arg}) variable containing the
            # files. if argument if not suppressed, some extra hoops are done
            # with remaps to ensure that files are also saved to the right
            # place. the main problem is that when jobs are submitted, the
            # directory structure is present in the submit node but not the
            # execute node, so when a job tries to create a file assuming the
            # directories are there, the job fails. this gets around the issue
            # by writing the files to the root directory then remaps them so
            # they get stored in the right place after the job completes and
            # files are transferred back
            for arg in node.outputs:
                if not arg.static:
                    if not arg.suppress:
                        basename = path.is_abs_or_url if self.transfer_files else False
                        nodevars[arg.condor_name] = arg.vars(basename=basename)
                    if self.transfer_files:
                        nodevars[f"output_{arg.condor_name}"] = arg.files(
                            basename=path.is_abs_or_url
                        )
                        nodevars[f"output_{arg.condor_name}_remap"] = arg.remaps()
            allvars.append(nodevars)

        return allvars

    def _update_submit_defaults(
        self, submit_description: Union[dict, htcondor.Submit]
    ) -> htcondor.Submit:
        # resolve executable path
        if os.path.exists(self.executable):
            executable = self.executable
        elif found_exec := shutil.which(self.executable):
            executable = found_exec
        else:
            warnings.warn(
                f"executable {self.executable} not found for layer {self.name}, "
                "this may be a failure mode during job submission",
                stacklevel=4,
            )
            executable = self.executable

        # add base submit opts + additional submit descriptors
        universe = submit_description.get("universe", self.universe)
        submit: Dict[str, Any] = {
            "universe": universe,
            "executable": executable,
            "arguments": self._arguments(),
            **submit_description,
        }

        # file submit opts
        if self.transfer_files:
            inputs = self._inputs()
            outputs = self._outputs()
            output_remaps = self._output_remaps()

            if inputs or outputs:
                submit.setdefault("should_transfer_files", "YES")
                submit.setdefault("when_to_transfer_output", "ON_SUCCESS")
                submit.setdefault("success_exit_code", 0)
                submit["preserve_relative_paths"] = True
            if inputs:
                submit["transfer_input_files"] = inputs
            if outputs:
                submit["transfer_output_files"] = outputs
                submit["transfer_output_remaps"] = f'"{output_remaps}"'

        # log submit opts
        submit.setdefault("output", "$(log_dir)/$(nodename)-$(cluster)-$(process).out")
        submit.setdefault("error", "$(log_dir)/$(nodename)-$(cluster)-$(process).err")

        # extra boilerplate submit opts
        submit.setdefault("notification", "never")

        return htcondor.Submit(format_submit_description(submit))


@dataclass
class Node:
    """Defines a single node (or job) in an HTCondor DAG.

    Stores both the arguments used within a job as well
    as capturing any inputs and outputs the job uses/creates.

    Parameters
    ----------
    arguments
        The arguments the node uses which aren't I/O related.
    inputs
        The arguments the node takes as inputs.
    outputs
        The arguments the node takes as outputs.
    variables
        Meta parameters that can be used within the submit description.

    """

    arguments: Union[Argument, Option, list] = field(default_factory=list)
    inputs: Union[Argument, Option, list] = field(default_factory=list)
    outputs: Union[Argument, Option, list] = field(default_factory=list)
    variables: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        if isinstance(self.arguments, Argument) or isinstance(self.arguments, Option):
            self.arguments = [self.arguments]
        if isinstance(self.inputs, Argument) or isinstance(self.inputs, Option):
            self.inputs = [self.inputs]
        if isinstance(self.outputs, Argument) or isinstance(self.outputs, Option):
            self.outputs = [self.outputs]

        # wrap string literals
        self.arguments = [_wrap_string_literal(arg) for arg in self.arguments]
        self.inputs = [_wrap_string_literal(arg) for arg in self.inputs]
        self.outputs = [_wrap_string_literal(arg) for arg in self.outputs]

    @property
    def requires(self) -> List[str]:
        """
        Returns
        -------
        list
            The inputs this node explicitly depends on to run.

        """
        assert isinstance(self.inputs, list)
        return list(
            itertools.chain(*[input_.args() for input_ in self.inputs if input_.track])
        )

    @property
    def provides(self) -> List[str]:
        """
        Returns
        -------
        list
            The outputs this node provides when it completes.

        """
        assert isinstance(self.outputs, list)
        return list(
            itertools.chain(*[output.args() for output in self.outputs if output.track])
        )


class HexFormatter(dags.SimpleFormatter):
    """A hex-based node formatter that produces names like LayerName:0000C."""

    def __init__(
        self, separator: str = ":", index_format: str = "{:05X}", offset: int = 0
    ) -> None:
        self.separator = separator
        self.index_format = index_format
        self.offset = offset

    def parse(self, node_name: str) -> Tuple[str, int]:
        layer, hex_index = node_name.split(self.separator)
        index = int(hex_index, 16)
        return layer, index - self.offset


def _wrap_string_literal(
    argument: Union[str, int, float, Argument, Option]
) -> Union[Literal, Argument, Option]:
    """Wraps a string literal, passing other arguments unchanged."""
    if isinstance(argument, Argument) or isinstance(argument, Option):
        return argument
    else:
        return Literal(argument)



import filecmp
from unittest import mock

import pytest

from .. import Argument, DAG, Option, Layer, Node


@mock.patch("shutil.which", side_effect=lambda x: f"/path/to/{x}")
def test_dag_generation(mock_which, shared_datadir, tmp_path):
    # create DAG
    dag_name = "my_dag"
    dag = DAG(dag_name)

    # define job requirements
    requirements = {
        "environment": {
            "OMP_NUM_THREADS": 1,
        },
        "getenv": [
            "HOME",
            "USER",
        ],
        "request_cpus": 1,
        "request_memory": 2000,
        "requirements": [
            "HAS_CVMFS_oasis_opensciencegrid_org=TRUE",
        ],
    }

    # create processing layer, add nodes
    process_layer = Layer("process_bins", submit_description=requirements)
    output_files = []
    for i in range(3):
        output_file = f"output_{i}.txt"
        bin_idx = [3 * j + i for j in range(3)]
        node = Node(
            arguments=[
                Argument("job-index", i),
                Option("verbose"),
                Option("bins", bin_idx),
            ],
            inputs=Option("input", "data.txt"),
            outputs=Argument("output", output_file),
        )

        process_layer.append(node)
        output_files.append(output_file)

        # check command
        cmd = process_layer.command(node)
        bin_cmd = Option("bins", bin_idx).vars()
        expected = (
            f"process_bins {i} --verbose {bin_cmd} --input data.txt {output_file}"
        )
        assert cmd == expected

    # add layer to DAG
    dag.attach(process_layer)

    # create combine layer, add node
    with pytest.warns(DeprecationWarning):
        combine_layer = Layer("combine_bins", requirements=requirements)
    combine_layer += Node(
        arguments=Option("verbose"),
        inputs=Argument("input", output_files),
        outputs=Argument("output", "combined.txt"),
    )

    # add layer to DAG
    dag.attach(combine_layer)

    # write DAG to disk
    dag.write(tmp_path, write_script=True)

    # compare contents of generated files
    dag_filename = f"{dag_name}.dag"
    script_filename = f"{dag_name}.sh"
    assert filecmp.cmp(
        tmp_path / dag_filename, shared_datadir / dag_filename
    ), f"contents for {dag_filename} does not match expected output"
    for job in ("process_bins", "combine_bins"):
        sub_filename = f"{job}.sub"
        sub_path = tmp_path / sub_filename
        assert filecmp.cmp(
            sub_path, shared_datadir / sub_filename
        ), f"contents for {sub_filename} does not match expected output"
    assert filecmp.cmp(
        tmp_path / script_filename, shared_datadir / script_filename
    ), f"contents for {script_filename} does not match expected output"


@mock.patch("shutil.which", side_effect=lambda x: f"/path/to/{x}")
@pytest.mark.parametrize(
    "name,base_path,transfer_files",
    [
        ["rel", "relative/path/to", True],
        ["rel", "relative/path/to", False],
        ["abs", "/absolute/path/to", True],
        ["abs", "/absolute/path/to", False],
        ["url", "osdf:///path/to", True],
    ],
)
def test_dag_file_paths(
    mock_which, name, base_path, transfer_files, shared_datadir, tmp_path
):
    transfer_path = "transfer" if transfer_files else "no_transfer"
    expected_datadir = shared_datadir / name / transfer_path

    # create DAG
    dag_name = "count_dag"
    dag = DAG(dag_name)

    # define job requirements
    options = {
        "request_cpus": 1,
        "request_memory": 2000,
    }

    # create partitioning layer
    partition_layer = Layer(
        "partition_by_key", submit_description=options, transfer_files=transfer_files
    )
    output_files = []
    for i in range(3):
        input_files = [f"/abs/dir/data_{3 * i + j}.txt" for j in range(3)]
        output_file = f"{base_path}/output_{i}.txt"
        partition_layer += Node(
            arguments=["-j", 4],
            inputs=Argument("data", input_files),
            outputs=Argument("partitioned", output_file),
        )
        output_files.append(output_file)
    dag.attach(partition_layer)

    # create aggregation layer
    agg_layer = Layer(
        "aggregate_by_key", submit_description=options, transfer_files=transfer_files
    )
    for output_file in output_files:
        agg_layer += Node(
            arguments=Option("aggregate", "max"),
            inputs=Argument("data", output_file),
            outputs=Argument("aggregated-data", output_file, suppress=True),
        )
    dag.attach(agg_layer)

    # create count layer
    count_layer = Layer(
        "count_keys", submit_description=options, transfer_files=transfer_files
    )
    count_layer += Node(
        arguments=Option("verbose"),
        inputs=Argument("input", output_files),
        outputs=Argument("output", "combined.txt"),
    )
    dag.attach(count_layer)

    # write DAG to disk
    dag.write(tmp_path, write_script=True)

    # compare contents of generated files
    dag_filename = f"{dag_name}.dag"
    script_filename = f"{dag_name}.sh"
    assert filecmp.cmp(
        tmp_path / dag_filename, expected_datadir / dag_filename
    ), f"contents for {dag_filename} does not match expected output"
    for job in ("partition_by_key", "aggregate_by_key", "count_keys"):
        sub_filename = f"{job}.sub"
        sub_path = tmp_path / sub_filename
        assert filecmp.cmp(
            sub_path, expected_datadir / sub_filename
        ), f"contents for {sub_filename} does not match expected output"
    assert filecmp.cmp(
        tmp_path / script_filename, expected_datadir / script_filename
    ), f"contents for {script_filename} does not match expected output"



# Copyright (C) 2020 Patrick Godwin
#
# This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
# If a copy of the MPL was not distributed with this file, You can obtain one at
# <https://mozilla.org/MPL/2.0/>.
#
# SPDX-License-Identifier: MPL-2.0

from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Callable, ClassVar, List, Optional, Union

from . import path

PROTECTED_CONDOR_VARS = {"input", "output", "rootdir"}


@dataclass
class Argument:
    """Defines a command-line argument (positional).

    This provides some extra functionality over defining command line
    argument explicitly, in addition to some extra parameters which
    sets how condor interprets how to handle them within the DAG
    and within submit descriptions.

    Parameters
    ----------
    name
        The option name. Since this is a positional argument, it is not
        used explicitly in the command, but is needed to define
        variable names within jobs.
    argument
        The positional argument value(s) used in a command.
    track
        Whether to track files defined here and used externally within
        jobs to determine parent-child relationships when nodes specify
        this option as an input or output. On by default.
    suppress
        Whether to hide this option. Used externally within jobs to
        determine whether to define job arguments. This is typically used
        when you want to track file I/O used by a job but isn't directly
        specified in their commands. Off by default.

    Examples
    --------
    >>> Argument("command", "run").vars()
    'run'

    >>> files = ["input_1.txt", "input_2.txt"]
    >>> Argument("input-files", files).vars()
    'input_1.txt input_2.txt'

    """

    name: str
    argument: Union[int, float, str, List]
    track: bool = True
    suppress: bool = False
    static: ClassVar[bool] = False
    _args: List[str] = field(init=False)

    def __post_init__(self) -> None:
        # check against list of protected condor names/characters,
        # rename condor variables name to avoid issues
        self.condor_name = self.name.replace("-", "_")
        if self.condor_name in PROTECTED_CONDOR_VARS:
            self.condor_name += "_"

        if isinstance(self.argument, str) or not isinstance(self.argument, Iterable):
            self.argument = [self.argument]
        self._args = [str(arg) for arg in self.argument]

    def args(self) -> List[str]:
        return self._args

    def vars(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        args = []
        for arg in self._args:
            args.append(path.normalize(arg, basename))
        return " ".join(args)

    def files(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        args = []
        for arg in self._args:
            args.append(path.normalize(arg, basename))
        return ",".join(args)

    def remaps(self) -> str:
        args = []
        for arg in self._args:
            if (normalized := path.normalize(arg, path.is_abs_or_url)) != arg:
                args.append(f"{normalized}={arg}")
        return ";".join(args)


@dataclass
class Option:
    """Defines a command-line option (long form).

    This provides some extra functionality over defining command line
    options explicitly, in addition to some extra parameters which
    sets how condor interprets how to handle them within the DAG
    and within submit descriptions.

    Parameters
    ----------
    name
        The option name to be used in a command.
    argument
        The argument value(s) used in a command.
    track
        Whether to track files defined here and used externally within
        jobs to determine parent-child relationships when nodes specify
        this option as an input or output. On by default.
    suppress
        Whether to hide this option. Used externally within jobs to
        determine whether to define job arguments. This is typically used
        when you want to track file I/O used by a job but isn't directly
        specified in their commands. Off by default.
    prefix
        The option prefix to use, e.g. for --verbose, -- is the prefix.
        Uses -- by default.

    Examples
    --------
    >>> Option("verbose").vars()
    '--verbose'

    >>> Option("input-type", "file").vars()
    '--input-type file'

    >>> Option("ifos", ["H1", "L1", "V1"]).vars()
    '--ifos H1 --ifos L1 --ifos V1'

    """

    name: str
    argument: Optional[Union[int, float, str, List]] = None
    track: Optional[bool] = True
    suppress: Optional[bool] = False
    prefix: Optional[str] = "--"
    static: ClassVar[bool] = False
    _args: List[str] = field(init=False)

    def __post_init__(self) -> None:
        # check against list of protected condor names/characters,
        # rename condor variables name to avoid issues
        self.condor_name = self.name.replace("-", "_")
        if self.condor_name in PROTECTED_CONDOR_VARS:
            self.condor_name += "_"

        if self.argument is not None:
            if isinstance(self.argument, str) or not isinstance(
                self.argument, Iterable
            ):
                self.argument = [self.argument]
            self._args = [str(arg) for arg in self.argument]

    def args(self) -> List[str]:
        return self._args

    def vars(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        if self.argument is None:
            return f"{self.prefix}{self.name}"
        args = []
        for arg in self._args:
            args.append(f"{self.prefix}{self.name} {path.normalize(arg, basename)}")
        return " ".join(args)

    def files(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        args = []
        for arg in self._args:
            args.append(path.normalize(arg, basename))
        return ",".join(args)

    def remaps(self) -> str:
        args = []
        for arg in self._args:
            if (normalized := path.normalize(arg, path.is_abs_or_url)) != arg:
                args.append(f"{normalized}={arg}")
        return ";".join(args)


@dataclass
class Literal:
    """Defines a command-line literal.

    This provides some extra functionality over defining command line
    argument explicitly, in addition to some extra parameters which
    sets how condor interprets how to handle them within the DAG
    and within submit descriptions.

    Parameters
    ----------
    argument
        The positional argument value(s) used in a command.
    track
        Whether to track files defined here and used externally within
        jobs to determine parent-child relationships when nodes specify
        this option as an input or output. On by default.

    Examples
    --------
    >>> Literal("run").vars()
    'run'

    >>> Literal("/path/to/input_1.txt").remaps()
    'input_1.txt=/path/to/input_1.txt'

    """

    argument: Union[int, float, str]
    track: bool = True
    static: ClassVar[bool] = True

    @property
    def name(self) -> str:
        return str(self.argument)

    def args(self) -> List[str]:
        return [self.name]

    def vars(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        return path.normalize(self.name, basename)

    def files(self, basename: Union[bool, Callable[[str], bool]] = False) -> str:
        return path.normalize(self.name, basename)

    def remaps(self) -> str:
        if (normalized := path.normalize(self.name, path.is_abs_or_url)) != self.name:
            return f"{normalized}={self.name}"
        else:
            return ""
